#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#include "Ctrl.h"

#define SEED 6834723

#ifndef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
	#define IS_SYNC
#endif

hit_tileNewType(float);
hit_ktileNewType(float);

double main_clock;
double exec_clock;

CTRL_KERNEL_CHAR(Mult, MANUAL, 2, LOCAL_SIZE_0, LOCAL_SIZE_1);


CTRL_KERNEL(Mult, OPENCL_GPU, KHitTile_float matrix_result, KHitTile_float matrix_a, KHitTile_float matrix_b, 
{
	__local float tile_a[LOCAL_SIZE_0][LOCAL_SIZE_1];
	__local float tile_b[LOCAL_SIZE_0][LOCAL_SIZE_1];

	int SIZE = matrix_result.card[1];

	int n_groups_x = get_num_groups(0);

	int local_size = get_local_size(0);

	int local_x = get_local_id(0);
	int local_y = get_local_id(1);
	
	float tmp = 0.0;
	int idx;
    
	for (int sub = 0; sub < n_groups_x; ++sub) {
		idx = thread_id.x * SIZE + sub * local_size + local_x;
		if (idx >= SIZE * SIZE) {
			tile_a[local_y][local_x] = 0;
		} else {
			tile_a[local_y][local_x] = hit_tileElemAtNoStride(matrix_a, 1, idx);
		}
		idx = (sub * local_size + local_y) * SIZE + thread_id.y;
		if (idx >= SIZE * SIZE) {
			tile_b[local_y][local_x] = 0;
		} else {
			tile_b[local_y][local_x] = hit_tileElemAtNoStride(matrix_b, 1, idx);
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	
		for (int k = 0; k < local_size; ++k) {
			tmp += tile_a[local_y][k] * tile_b[k][local_x];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}
	if (thread_id.x < SIZE && thread_id.y < SIZE) {
		hit_tileElemAtNoStride(matrix_result, 2, thread_id.x, thread_id.y) = tmp;
	}
});

CTRL_KERNEL_PROTO( Mult,
    1, OPENCL_GPU,
    3, OUT, HitTile_float, matrix_result, IN, HitTile_float, matrix_a, IN, HitTile_float, matrix_b
);


int main(int argc, char *argv[])
{
	main_clock = omp_get_wtime();

	if (argc != 5) {
		fprintf(stderr, "\nUsage: %s <size> <n_iter> <platform> <device>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	int SIZE = atoi(argv[1]);
	int N_ITER = atoi(argv[2]);
	int PLATFORM = atoi(argv[3]);
	int DEVICE = atoi(argv[4]);


	HitTile_float matrix_a, matrix_b, matrix_c;
	hit_tileDomain(&matrix_a, float, 2, SIZE, SIZE);
	hit_tileDomain(&matrix_b, float, 2, SIZE, SIZE);
    hit_tileDomain(&matrix_c, float, 2, SIZE, SIZE);
	
	#ifdef IS_SYNC
		double *p_res = (double *)malloc(N_ITER * sizeof(double));
		double *p_sum = (double *)malloc(N_ITER * sizeof(double));
		HitTile_float matrix_tmp;
		hit_tileDomain(&matrix_tmp, float, 2, SIZE, SIZE);
		hit_tileAlloc(&matrix_tmp);
	#endif

	Ctrl_Thread threads;
	Ctrl_ThreadInit(threads, 2, SIZE, SIZE);

	__ctrl_block__(1)
	{
		Ctrl ctrl;

		#ifdef IS_SYNC
			Ctrl_Create(&ctrl, CTRL_TYPE_OPENCL_GPU, CTRL_POLICY_SYNC, DEVICE, &PLATFORM);
		#else
			Ctrl_Create(&ctrl, CTRL_TYPE_OPENCL_GPU, CTRL_POLICY_ASYNC, DEVICE, &PLATFORM);
		#endif

		Ctrl_AllocPinned(&ctrl, (HitTile *)&matrix_a);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&matrix_b);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&matrix_c);

		for (int i = 0; i < SIZE; i++) {
			for (int j = 0; j < SIZE; j++) {
				hit_tileElemAt(matrix_a, 2, i, j) = 0.001;
				hit_tileElemAt(matrix_b, 2, i, j) = 0.001;
				hit_tileElemAt(matrix_c, 2, i, j) = 0;
			}
		}

		Ctrl_Attach(&ctrl, (HitTile *)&matrix_a);
		Ctrl_Attach(&ctrl, (HitTile *)&matrix_b);
		Ctrl_Attach(&ctrl, (HitTile *)&matrix_c);

		Ctrl_GlobalSync(&ctrl);
		exec_clock = omp_get_wtime();

        for (int i = 0; i < N_ITER; i++) {
			if ((i % 2) == 0) {
				Ctrl_Launch(&ctrl, Mult, threads, matrix_c, matrix_a, matrix_b);
				Ctrl_MoveFrom(&ctrl, (HitTile *)&matrix_c);
				#ifdef IS_SYNC
					double minimo = hit_tileElemAt(matrix_c, 2, 0, 0);
					double maximo = hit_tileElemAt(matrix_c, 2, 0, 0);

					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							if (minimo > hit_tileElemAt(matrix_c, 2, j, k)){
								minimo = hit_tileElemAt(matrix_c, 2, j, k);
							}
							if (maximo < hit_tileElemAt(matrix_c, 2, j, k)){
								maximo = hit_tileElemAt(matrix_c, 2, j, k);
							}
						}
					}

					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							hit_tileElemAt(matrix_c, 2, j, k) = hit_tileElemAt(matrix_c, 2, j, k) - minimo;
							hit_tileElemAt(matrix_c, 2, j, k) = hit_tileElemAt(matrix_c, 2, j, k) / maximo; 
						}
					}

					p_sum[i] = 0;
					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							p_sum[i] += pow(hit_tileElemAt(matrix_c, 2, j, k), 2);
						}
					}
					p_res[i] = sqrt(p_sum[i]);

					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							hit_tileElemAt(matrix_c, 2, j, k) = hit_tileElemAt(matrix_c, 2, j, k) / p_res[i];
						}
					}				
					memcpy(matrix_tmp.data, matrix_c.data, sizeof(float) * SIZE * SIZE);
				#endif
			} else {
				Ctrl_Launch(&ctrl, Mult, threads, matrix_b, matrix_a, matrix_c);
				Ctrl_MoveFrom(&ctrl, (HitTile *)&matrix_b);
				#ifdef IS_SYNC
					double minimo = hit_tileElemAt(matrix_b, 2, 0, 0);
					double maximo = hit_tileElemAt(matrix_b, 2, 0, 0);

					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							if (minimo > hit_tileElemAt(matrix_b, 2, j, k)){
								minimo = hit_tileElemAt(matrix_b, 2, j, k);
							}
							if (maximo < hit_tileElemAt(matrix_b, 2, j, k)){
								maximo = hit_tileElemAt(matrix_b, 2, j, k);
							}
						}
					}

					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							hit_tileElemAt(matrix_b, 2, j, k) = hit_tileElemAt(matrix_b, 2, j, k) - minimo;
							hit_tileElemAt(matrix_b, 2, j, k) = hit_tileElemAt(matrix_b, 2, j, k) / maximo; 
						}
					}

					p_sum[i] = 0;
					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							p_sum[i] += pow(hit_tileElemAt(matrix_b, 2, j, k), 2);
						}
					}
					p_res[i] = sqrt(p_sum[i]);

					for (int j = 0; j < SIZE; j++) {
						for (int k = 0; k < SIZE; k++) {
							hit_tileElemAt(matrix_b, 2, j, k) = hit_tileElemAt(matrix_b, 2, j, k) / p_res[i];
						}
					}				
					memcpy(matrix_tmp.data, matrix_b.data, sizeof(float) * SIZE * SIZE);	
				#endif
			}
        }

		Ctrl_GlobalSync(&ctrl);
		exec_clock = omp_get_wtime() - exec_clock;

		Ctrl_Detach(&ctrl, (HitTile *)&matrix_a);
		Ctrl_Detach(&ctrl, (HitTile *)&matrix_b);
        Ctrl_Detach(&ctrl, (HitTile *)&matrix_c);

		#ifndef IS_SYNC
			double resultado = 0, suma = 0;
			for (int i = 0; i < SIZE; i++) {
				for (int j = 0; j < SIZE; j++) {
					suma += pow(hit_tileElemAt(((N_ITER % 2) == 0)? matrix_b : matrix_c, 2, i, j), 2);
				}
			}
			resultado = sqrt(suma);
		#endif

		printf("\n ----------------------- NORM ----------------------- \n\n"); fflush(stdout);
		#ifdef IS_SYNC
			for (int i = 0; i < N_ITER; i++) {
				printf(" iter: %d, sum: %lf, res: %lf\n", i, p_sum[i], p_res[i]); fflush(stdout);
			}
		#else
			printf(" Acumulated sum: %lf \n", suma); fflush(stdout);
			printf(" Result: %lf \n", resultado); fflush(stdout);
		#endif
		printf("\n ---------------------------------------------------- \n"); fflush(stdout);

		Ctrl_FreePinned(&ctrl, (HitTile *)&matrix_a);
		Ctrl_FreePinned(&ctrl, (HitTile *)&matrix_b);
		Ctrl_FreePinned(&ctrl, (HitTile *)&matrix_c);
			
		Ctrl_Destroy(&ctrl);
	}
	
	#ifdef IS_SYNC
		free(p_sum);
		free(p_res);
		hit_tileFree(matrix_tmp);
	#endif


	main_clock = omp_get_wtime() - main_clock;

	printf("\n ----------------------- TIME ----------------------- \n\n");
	printf(" Clock main: %lf\n", main_clock);
	printf(" Clock exec: %lf\n", exec_clock);
	printf("\n ---------------------------------------------------- \n");

	return 0;
}
