#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#include "Ctrl.h"

#define SEED 6834723

#ifndef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
	#define IS_SYNC
#endif

hit_tileNewType(float);
hit_ktileNewType(float);

double main_clock;
double exec_clock;

CTRL_KERNEL_CHAR(Cell_Autom, MANUAL, 2, LOCAL_SIZE_0, LOCAL_SIZE_1);
// CTRL_KERNEL_SHMEM


CTRL_KERNEL(Cell_Autom, OPENCL_GPU, KHitTile_float dest, KHitTile_float src,
{
	int SIZE = dest.card[0];
    thread_id.x++;
    thread_id.y++;
    hit_tileElemAtNoStride(dest, 2, thread_id.x, thread_id.y) = ( 
            hit_tileElemAtNoStride(src, 2, thread_id.x - 1, thread_id.y    ) +
            hit_tileElemAtNoStride(src, 2, thread_id.x + 1, thread_id.y    ) +
            hit_tileElemAtNoStride(src, 2, thread_id.x    , thread_id.y - 1) +
            hit_tileElemAtNoStride(src, 2, thread_id.x    , thread_id.y + 1) ) / 4.0;
});


CTRL_KERNEL_PROTO( Cell_Autom,
    1, OPENCL_GPU,
    2, OUT, HitTile_float, dest, IN, HitTile_float, src
);


int main(int argc, char *argv[])
{
	main_clock = omp_get_wtime();

	if (argc != 5) {
		fprintf(stderr, "\nUsage: %s <size> <n_iter> <platform> <device>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	int SIZE = atoi(argv[1]);
	int N_ITER = atoi(argv[2]);
	int PLATFORM = atoi(argv[3]);
	int DEVICE = atoi(argv[4]);

	HitTile_float matrix_1;
	HitTile_float matrix_2;
	
	hit_tileDomain(&matrix_1, float, 2, SIZE, SIZE);
    hit_tileDomain(&matrix_2, float, 2, SIZE, SIZE);
	
	#ifdef IS_SYNC
		HitTile_float matrix_res;
		hit_tileDomain(&matrix_res, float, 2, SIZE, SIZE);
	#endif

	Ctrl_Thread threads;
	Ctrl_ThreadInit(threads, 2, SIZE - 2, SIZE - 2);

	__ctrl_block__(1)
	{
		Ctrl ctrl;

		#ifdef IS_SYNC
			Ctrl_Create(&ctrl, CTRL_TYPE_OPENCL_GPU, CTRL_POLICY_SYNC, DEVICE, &PLATFORM);
		#else
			Ctrl_Create(&ctrl, CTRL_TYPE_OPENCL_GPU, CTRL_POLICY_ASYNC, DEVICE, &PLATFORM);
		#endif

		Ctrl_AllocPinned(&ctrl, (HitTile *)&matrix_1);
		Ctrl_AllocPinned(&ctrl, (HitTile *)&matrix_2);

		#ifdef IS_SYNC
			hit_tileAlloc(&matrix_res);
		#endif

		for (int i = 0; i < SIZE; i++) {
            for (int j = 0; j < SIZE; j++) {
                hit_tileElemAt(matrix_1, 2, i, j) = 1.1;
                hit_tileElemAt(matrix_2, 2, i, j) = 0.0;
            }
	    }

        for (int j = 0; j < SIZE; j++){
            hit_tileElemAt(matrix_1, 2, 0, j) = 2.1;
            hit_tileElemAt(matrix_1, 2, SIZE - 1, j) = 2.1;
        }
        for (int i = 0; i < SIZE; i++) {
            hit_tileElemAt(matrix_1, 2, i, 0) = 3.2;
            hit_tileElemAt(matrix_1, 2, i, SIZE - 1) = 4.1;
        }


		Ctrl_Attach(&ctrl, (HitTile *)&matrix_1);
		Ctrl_Attach(&ctrl, (HitTile *)&matrix_2);

		Ctrl_GlobalSync(&ctrl);
		exec_clock = omp_get_wtime();

        for (int i = 1; i <= N_ITER; i++) {
			if ((i % 2) == 0) {
				Ctrl_Launch(&ctrl, Cell_Autom, threads, matrix_1, matrix_2);
			} else {
				Ctrl_Launch(&ctrl, Cell_Autom, threads, matrix_2, matrix_1);
			}

			if ((i % 10) == 0) {
				Ctrl_MoveFrom(&ctrl, (HitTile *)&matrix_1);
				
				#ifdef IS_SYNC 
					memcpy(matrix_res.data, matrix_1.data, sizeof(float)*SIZE*SIZE );
				#endif
			}
        }

		Ctrl_GlobalSync(&ctrl);
		exec_clock = omp_get_wtime() - exec_clock;

		double resultado = 0, suma = 0;
		for (int i = 0; i < SIZE; i++) {
			for (int j = 0; j < SIZE; j++) {
				#ifdef IS_SYNC
					suma += pow(hit_tileElemAt(matrix_res, 2, i, j), 2);
				#else
					suma += pow(hit_tileElemAt(matrix_1 , 2, i, j), 2);
				#endif
			}
		}
		resultado = sqrt(suma);

		Ctrl_Detach(&ctrl, (HitTile *)&matrix_1);
        	Ctrl_Detach(&ctrl, (HitTile *)&matrix_2);


		printf("\n ----------------------- NORM ----------------------- \n\n");
		fflush(stdout);
		printf(" Acumulated sum: %lf \n", suma);
		fflush(stdout);
		printf(" Result: %lf \n", resultado);
		fflush(stdout);
		printf("\n ---------------------------------------------------- \n");
		fflush(stdout);

		Ctrl_FreePinned(&ctrl, (HitTile *)&matrix_1);
		Ctrl_FreePinned(&ctrl, (HitTile *)&matrix_2);

		#ifdef IS_SYNC
			hit_tileFree(matrix_res);
		#endif

		Ctrl_Destroy(&ctrl);
	}

	main_clock = omp_get_wtime() - main_clock;

	printf("\n ----------------------- TIME ----------------------- \n\n");
	printf(" Clock main: %lf\n", main_clock);
	printf(" Clock exec: %lf\n", exec_clock);
	printf("\n ---------------------------------------------------- \n");

	return 0;
}
